from .data_loader_base import DataLoaderQDRBase
from dataclasses import dataclass
from typing import Iterator

@dataclass
class Signal1MQueryInstance:
    id : str
    text : str

@dataclass
class Signal1MDocInstance:
    id : str
    text : str

@dataclass
class Signal1MQrelInstance:
    qid : str
    did : str
    score : int

class Signal1MDataLoader(DataLoaderQDRBase):
    def __init__(self, *args, **kwargs):
        super().__init__('beir/signal1m', *args, **kwargs)

    def make_query_instance(self, **kwargs) -> Signal1MQueryInstance:
        return Signal1MQueryInstance(**kwargs)

    def make_doc_instance(self, **kwargs) -> Signal1MDocInstance:
        return Signal1MDocInstance(**kwargs)

    def make_qrel_instance(self, **kwargs) -> Signal1MQrelInstance:
        return Signal1MQrelInstance(**kwargs)

    def __str__(self) -> str:
        return 'Signal1MDataLoader()'

    def __repr__(self) -> str:
        return 'Signal1MDataLoader()'



from .data_loader_base import DataLoaderQDRBase
from dataclasses import dataclass
from typing import Iterator

@dataclass
class Robust04QueryInstance:
    id : str
    text : str

@dataclass
class Robust04DocInstance:
    id : str
    text : str

@dataclass
class Robust04QrelInstance:
    qid : str
    did : str
    score : int

class Robust04DataLoader(DataLoaderQDRBase):
    def __init__(self, *args, **kwargs):
        super().__init__('beir/robust04', *args, **kwargs)

    def make_query_instance(self, **kwargs) -> Robust04QueryInstance:
        return Robust04QueryInstance(**kwargs)

    def make_doc_instance(self, **kwargs) -> Robust04DocInstance:
        return Robust04DocInstance(**kwargs)

    def make_qrel_instance(self, **kwargs) -> Robust04QrelInstance:
        return Robust04QrelInstance(**kwargs)

    def __str__(self) -> str:
        return 'Robust04DataLoader()'

    def __repr__(self) -> str:
        return 'Robust04DataLoader()'



from .data_loader_base import DataLoaderQDRBase
from dataclasses import dataclass
from typing import Iterator

@dataclass
class SciFactQueryInstance:
    id : str
    text : str

@dataclass
class SciFactDocInstance:
    id : str
    text : str
    title : str

@dataclass
class SciFactQrelInstance:
    qid : str
    did : str
    score : int

class SciFactDataLoader(DataLoaderQDRBase):
    def __init__(self, *args, **kwargs):
        super().__init__('beir/scifact', *args, **kwargs)

    def make_query_instance(self, **kwargs) -> SciFactQueryInstance:
        return SciFactQueryInstance(**kwargs)

    def make_doc_instance(self, **kwargs) -> SciFactDocInstance:
        return SciFactDocInstance(**kwargs)

    def make_qrel_instance(self, **kwargs) -> SciFactQrelInstance:
        return SciFactQrelInstance(**kwargs)

    def __str__(self) -> str:
        return 'SciFactDataLoader()'

    def __repr__(self) -> str:
        return 'SciFactDataLoader()'



from .data_loader_base import DataLoaderQDRBase
from dataclasses import dataclass
from typing import Iterator

@dataclass
class ClimateFEVERQueryInstance:
    id : str
    text : str

@dataclass
class ClimateFEVERDocInstance:
    id : str
    text : str
    title : str

@dataclass
class ClimateFEVERQrelInstance:
    qid : str
    did : str
    score : int

class ClimateFEVERDataLoader(DataLoaderQDRBase):
    def __init__(self, *args, **kwargs):
        super().__init__('beir/climate-fever', *args, **kwargs)

    def make_query_instance(self, **kwargs) -> ClimateFEVERQueryInstance:
        return ClimateFEVERQueryInstance(**kwargs)

    def make_doc_instance(self, **kwargs) -> ClimateFEVERDocInstance:
        return ClimateFEVERDocInstance(**kwargs)

    def make_qrel_instance(self, **kwargs) -> ClimateFEVERQrelInstance:
        return ClimateFEVERQrelInstance(**kwargs)

    def __str__(self) -> str:
        return 'ClimateFEVERDataLoader()'

    def __repr__(self) -> str:
        return 'ClimateFEVERDataLoader()'



CHO = ['ㄱ', 'ㄲ', 'ㄴ', 'ㄷ', 'ㄸ', 'ㄹ', 'ㅁ', 'ㅂ', 'ㅃ', 'ㅅ', 'ㅆ', 'ㅇ', 'ㅈ', 'ㅉ', 'ㅊ', 'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ']
JUNG = ['ㅏ', 'ㅐ', 'ㅑ', 'ㅒ', 'ㅓ', 'ㅔ', 'ㅕ', 'ㅖ', 'ㅗ', 'ㅘ', 'ㅙ', 'ㅚ', 'ㅛ', 'ㅜ', 'ㅝ', 'ㅞ', 'ㅟ', 'ㅠ', 'ㅡ', 'ㅢ', 'ㅣ']
JONG = ['', 'ㄱ', 'ㄲ', 'ㄳ', 'ㄴ', 'ㄵ', 'ㄶ', 'ㄷ', 'ㄹ', 'ㄺ', 'ㄻ', 'ㄼ', 'ㄽ', 'ㄾ', 'ㄿ', 'ㅀ', 'ㅁ', 'ㅂ', 'ㅄ', 'ㅅ', 'ㅆ', 'ㅇ', 'ㅈ', 'ㅊ', 'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ']

def split_korean_jaso(text : str, remove_space : bool = False) -> list[str]:
    """Splits Korean text into individual Jaso (initial, medial, and final consonants).

    Args:
        text (str): The input string containing Korean characters to be split into Jaso.
        remove_space (bool, optional): If True, spaces will be removed from the output. Defaults to False.

    Returns:
        list[str]: A list of strings, where each Korean character is split into its constituent Jaso.
    """
    split_text = []
    for char in text:
        if remove_space and char == ' ':
            continue
        if '가' <= char <= '힣':
            char_code = ord(char) - 44032
            jong = char_code % 28
            jung = ((char_code - jong) // 28) % 21
            cho = ((char_code - jong) // 28) // 21
            split_text.append(CHO[cho])
            split_text.append(JUNG[jung])
            if jong != 0:
                split_text.append(JONG[jong])
        else:
            split_text.append(char)
    return split_text



from .data_loader_base import DataLoaderQDRBase
from dataclasses import dataclass
from typing import Iterator

@dataclass
class SciDocsQueryInstance:
    id : str
    text : str
    authors : list[str]
    year : int
    cited_by : list[str]
    references : list[str]

@dataclass
class SciDocsDocInstance:
    id : str
    text : str
    title : str
    authors : list[str]
    year : int
    cited_by : list[str]
    references : list[str]

@dataclass
class SciDocsQrelInstance:
    qid : str
    did : str
    score : int

class SciDocsDataLoader(DataLoaderQDRBase):
    def __init__(self, *args, **kwargs):
        super().__init__('beir/scidocs', *args, **kwargs)

    def make_query_instance(self, **kwargs) -> SciDocsQueryInstance:
        return SciDocsQueryInstance(**kwargs)

    def make_doc_instance(self, **kwargs) -> SciDocsDocInstance:
        return SciDocsDocInstance(**kwargs)

    def make_qrel_instance(self, **kwargs) -> SciDocsQrelInstance:
        return SciDocsQrelInstance(**kwargs)

    def __str__(self) -> str:
        return 'SciDocsDataLoader()'

    def __repr__(self) -> str:
        return 'SciDocsDataLoader()'



from .data_loader_base import DataLoaderQDRBase
from dataclasses import dataclass
from typing import Iterator

@dataclass
class QuoraQueryInstance:
    id : str
    text : str

@dataclass
class QuoraDocInstance:
    id : str
    text : str

@dataclass
class QuoraQrelInstance:
    qid : str
    did : str
    score : int

class QuoraDataLoader(DataLoaderQDRBase):
    def __init__(self, *args, **kwargs):
        super().__init__('beir/quora', *args, **kwargs)

    def make_query_instance(self, **kwargs) -> QuoraQueryInstance:
        return QuoraQueryInstance(**kwargs)

    def make_doc_instance(self, **kwargs) -> QuoraDocInstance:
        return QuoraDocInstance(**kwargs)

    def make_qrel_instance(self, **kwargs) -> QuoraQrelInstance:
        return QuoraQrelInstance(**kwargs)

    def __str__(self) -> str:
        return 'QuoraDataLoader()'

    def __repr__(self) -> str:
        return 'QuoraDataLoader()'



from .data_loader_base import DataLoaderQDRBase
from dataclasses import dataclass
from typing import Iterator

@dataclass
class Touchev2QueryInstance:
    id : str
    text : str
    description : str
    narrative : str

@dataclass
class Touchev2DocInstance:
    id : str
    text : str
    title : str
    stance : str
    url : str

@dataclass
class Touchev2QrelInstance:
    qid : str
    did : str
    score : int

class Touchev2DataLoader(DataLoaderQDRBase):
    def __init__(self, *args, **kwargs):
        super().__init__('beir/touche-v2', *args, **kwargs)

    def make_query_instance(self, **kwargs) -> Touchev2QueryInstance:
        return Touchev2QueryInstance(**kwargs)

    def make_doc_instance(self, **kwargs) -> Touchev2DocInstance:
        return Touchev2DocInstance(**kwargs)

    def make_qrel_instance(self, **kwargs) -> Touchev2QrelInstance:
        return Touchev2QrelInstance(**kwargs)

    def __str__(self) -> str:
        return 'Touchev2DataLoader()'

    def __repr__(self) -> str:
        return 'Touchev2DataLoader()'



from ..utils.corpus_reader import CorpusReader
from dataclasses import asdict
from tqdm import tqdm
import glob
import os
import requests
import shutil
import tempfile
import urllib.parse

class DataLoaderBase:
    """Base class for DataLoader
    """
    def __init__(self, dataset_name : str, force_download : bool = False):
        """Constructor for DataLoaderBase

        Args:
            dataset_name (str): Name of the dataset to load.
            force_download (bool, optional): Whether to force download the dataset. Defaults to False.
        """
        self.dataset_name = dataset_name
        self.download_urls = self._fetch_download_urls(dataset_name)
        self.data_dir = os.path.join(tempfile.gettempdir(), 'hamu_tool', 'dataset', self.dataset_name)
        if os.path.exists(self.data_dir):
            if not force_download:
                return
            shutil.rmtree(self.data_dir)
        os.makedirs(self.data_dir)
        print(f'Downloading dataset [{self.dataset_name}] ...')
        for url in self.download_urls:
            self._download_from_url(url, self.data_dir)

    def _fetch_download_urls(self, dataset_name : str) -> list[str]:
        """Fetch the download urls for the given dataset.

        Args:
            dataset_name (str): Name of the dataset to fetch download urls.

        Raises:
            Exception: If failed to fetch download urls for the dataset.

        Returns:
            list[str]: List of download urls for the dataset.
        """
        res = requests.get(f'http://research.hamu.me/dataset/api/get_download_url/{urllib.parse.quote(dataset_name, safe="")}/')
        if res.status_code == 200:
            data = res.json()
            download_urls = data['download_url'].split('\n')
            if len(download_urls) == 0 or download_urls[0] == '':
                raise Exception(f'No download urls found for the dataset [{dataset_name}]')
            return download_urls
        else:
            raise Exception('Failed to fetch download urls')

    def _download_from_url(self, url : str, data_dir : str) -> None:
        """Download the dataset from the given url.

        Args:
            url (str): URL to download the dataset.
            data_dir (str): Directory to save the downloaded dataset.

        Raises:
            Exception: If failed to download the dataset.
        """
        res = requests.get(url, stream=True)
        if res.status_code == 200:
            filename = res.headers.get('content-disposition').split('filename=')
            if len(filename) > 1:
                filename = filename[1]
            else:
                filename = res.headers.get('content-disposition').split('filename*=')[-1].split("''")[-1]
            total = int(res.headers.get('content-length', 0))
            pbar = tqdm(total=total, desc=filename, unit='B', unit_scale=True, unit_divisor=1024)
            data_path = os.path.join(data_dir, filename)
            with open(data_path, 'wb') as fp:
                for data in res.iter_content(chunk_size=1024):
                    pbar.update(len(data))
                    fp.write(data)
                # fp.write(res.content)
        else:
            raise Exception('Failed to download dataset')

class DataLoaderQDRBase(DataLoaderBase):
    """Base class for DataLoaderQDR
    """
    def __init__(self, dataset_name : str, *args, **kwargs):
        """Constructor for DataLoaderQDRBase

        Args:
            dataset_name (str): Name of the dataset to load.
        """
        super().__init__(dataset_name, *args, **kwargs)
        self.reader_doc = CorpusReader(f'{self.data_dir}/doc.idx')
        self.reader_query = CorpusReader(f'{self.data_dir}/query.idx')
        self.qrel = {}
        self.qrel_list = {}
        self.drel = {}
        self.qid_list = {}
        self.did_list = {}
        qrel_paths = glob.glob(f'{self.data_dir}/qrel.*.tsv')
        for qrel_path in qrel_paths:
            mode = qrel_path.split('.')[-2]
            self.qrel[mode] = {}
            self.qrel_list[mode] = []
            self.drel[mode] = {}
            self.qid_list[mode] = []
            self.qid_set = set()
            self.did_list[mode] = []
            self.did_set = set()
            with open(qrel_path, 'r', encoding='utf-8') as fp:
                for line in fp:
                    qid, _, did, score = line.strip().split()
                    if qid not in self.qrel[mode]:
                        self.qrel[mode][qid] = []
                    self.qrel[mode][qid].append((did, int(score)))
                    self.qrel_list[mode].append((qid, did, int(score)))
                    if did not in self.drel[mode]:
                        self.drel[mode][did] = []
                    self.drel[mode][did].append((qid, int(score)))
                    if qid not in self.qid_set:
                        self.qid_set.add(qid)
                        self.qid_list[mode].append(qid)
                    if did not in self.did_set:
                        self.did_set.add(did)
                        self.did_list[mode].append(did)

    def total_docs(self, mode : str = None) -> int:
        """Total number of documents in the dataset.

        Args:
            mode (str, optional): Mode of the dataset. Defaults to None.

        Returns:
            int: Total number of documents in the dataset.
        """
        if not mode:
            size = len(self.reader_doc)
        else:
            size = len(self.did_list[mode])
        return size

    def get_did(self, idx : int, mode : str = None) -> str:
        """Fetch the document ID by its index.

        Args:
            idx (int): The index of the document.
            mode (str, optional): Mode of the dataset. Defaults to None.

        Returns:
            str: The fetched document ID.
        """
        if not mode:
            did = self.reader_doc.idx_list[idx]
        else:
            did = self.did_list[mode][idx]
        return did

    def total_queries(self, mode : str = None) -> int:
        """Total number of queries in the dataset.

        Args:
            mode (str, optional): Mode of the dataset. Defaults to None.

        Returns:
            int: Total number of queries in the dataset.
        """
        if not mode:
            size = len(self.reader_query)
        else:
            size = len(self.qid_list[mode])
        return size

    def get_qid(self, idx : int, mode : str = None) -> str:
        """Fetch the query ID by its index.

        Args:
            idx (int): The index of the query.

        Returns:
            str: The fetched query ID.
        """
        if not mode:
            qid = self.reader_query.idx_list[idx]
        else:
            qid = self.qid_list[mode][idx]
        return qid

    def total_qrels(self, mode : str) -> int:
        """Total number of qrels in the dataset.

        Args:
            mode (str): Mode of the dataset.

        Returns:
            int: Total number of qrels in the dataset.
        """
        return len(self.qrel_list[mode])

    def make_query_instance(self, **kwargs):
        """Make a query instance.

        Args:
            **kwargs: Keyword arguments for the query instance.

        Returns:
            QueryInstance: The query instance dataclass.
        """
        raise NotImplementedError

    def make_doc_instance(self, **kwargs):
        """Make a document instance.

        Args:
            **kwargs: Keyword arguments for the document instance.

        Returns:
            DocInstance: The document instance dataclass.
        """
        raise NotImplementedError

    def make_qrel_instance(self, **kwargs):
        """Make a qrel instance.

        Args:
            **kwargs: Keyword arguments for the qrel instance.

        Returns:
            QrelInstance: The qrel instance dataclass.
        """
        raise NotImplementedError

    def get_query(self, qid : str | int, mode : str = None):
        """Fetch the query instance by its ID.

        Args:
            qid (str | int): The query ID or index.
            mode (str, optional): Mode of the dataset. Defaults to None.

        Returns:
            QueryInstance: The fetched query instance.
        """
        if isinstance(qid, int):
            qid = self.get_qid(qid, mode)
        query = self.reader_query[qid]
        instance = self.make_query_instance(**asdict(query))
        return instance

    def get_queries(self, mode : str = None):
        """Fetch all query instances.

        Args:
            mode (str, optional): Mode of the dataset. Defaults to None.

        Yields:
            QueryInstance: The fetched query instance.
        """
        if not mode:
            for query in self.reader_query:
                instance = self.make_query_instance(**asdict(query))
                yield instance
        else:
            for qid in self.qid_list[mode]:
                query = self.reader_query[qid]
                instance = self.make_query_instance(**asdict(query))
                yield instance

    def get_doc(self, did : str | int, mode : str = None):
        """Fetch the document instance by its ID.

        Args:
            did (str | int): The document ID or index.
            mode (str, optional): Mode of the dataset. Defaults to None.

        Returns:
            DocInstance: The fetched document instance.
        """
        if isinstance(did, int):
            did = self.get_did(did, mode)
        doc = self.reader_doc[did]
        instance = self.make_doc_instance(**asdict(doc))
        return instance

    def get_docs(self, mode : str = None):
        """Fetch all document instances.

        Args:
            mode (str, optional): Mode of the dataset. Defaults to None.

        Yields:
            DocInstance: The fetched document instance.
        """
        if not mode:
            for doc in self.reader_doc:
                instance = self.make_doc_instance(**asdict(doc))
                yield instance
        else:
            for did in self.did_list[mode]:
                doc = self.reader_doc[did]
                instance = self.make_doc_instance(**asdict(doc))
                yield instance

    def get_qrel(self, mode : str, qid : str):
        """Fetch the qrel instances by query ID.

        Args:
            mode (str): Mode of the dataset.
            qid (str): The query ID.

        Raises:
            KeyError: If qrel for the query ID not found.

        Returns:
            list[QrelInstance]: List of fetched qrel instances.
        """
        if qid not in self.qrel[mode]:
            raise KeyError(f'Qrel for query [{qid}] not found')
        instances = []
        for did, score in self.qrel[mode][qid]:
            instances.append(self.make_qrel_instance(qid=qid, did=did, score=score))
        return instances

    def get_qrels(self, mode : str):
        """Fetch all qrel instances.

        Args:
            mode (str): Mode of the dataset.

        Yields:
            QrelInstance: The fetched qrel instance.
        """
        for qid, did, score in self.qrel_list[mode]:
            instance = self.make_qrel_instance(qid=qid, did=did, score=score)
            yield instance

    def get_drel(self, mode : str, did : str):
        """Fetch the drel instances by document ID.

        Args:
            mode (str): Mode of the dataset.
            did (str): The document ID.

        Raises:
            KeyError: If drel for the document ID not found.

        Returns:
            list[QrelInstance]: List of fetched drel instances.
        """
        if did not in self.drel[mode]:
            raise KeyError(f'Drel for document [{did}] not found')
        instances = []
        for qid, score in self.drel[mode][did]:
            instances.append(self.make_qrel_instance(qid=qid, did=did, score=score))
        return instances

    def get_drels(self, mode : str):
        """Fetch all drel instances.

        Args:
            mode (str): Mode of the dataset.

        Returns:
            list[QrelInstance]: List of fetched drel instances.
        """
        return self.get_qrels(mode)



from .data_loader_base import DataLoaderQDRBase
from dataclasses import dataclass
from typing import Iterator

@dataclass
class ToucheQueryInstance:
    id : str
    text : str
    description : str
    narrative : str

@dataclass
class ToucheDocInstance:
    id : str
    text : str
    title : str
    stance : str
    url : str

@dataclass
class ToucheQrelInstance:
    qid : str
    did : str
    score : int

class ToucheDataLoader(DataLoaderQDRBase):
    def __init__(self, *args, **kwargs):
        super().__init__('beir/touche', *args, **kwargs)

    def make_query_instance(self, **kwargs) -> ToucheQueryInstance:
        return ToucheQueryInstance(**kwargs)

    def make_doc_instance(self, **kwargs) -> ToucheDocInstance:
        return ToucheDocInstance(**kwargs)

    def make_qrel_instance(self, **kwargs) -> ToucheQrelInstance:
        return ToucheQrelInstance(**kwargs)

    def __str__(self) -> str:
        return 'ToucheDataLoader()'

    def __repr__(self) -> str:
        return 'ToucheDataLoader()'



"""This module defines a CorpusReader class for efficient access to documents in a corpus using an index.
The CorpusReader uses memory-mapped file support for random access to large files or streams.
This module is compatible with Python 3.8 and above due to the use of the '|' (pipe) operator for type hinting.
The CorpusReader class provides methods to:
- Initialize the corpus reader using a pre-built index file.
- Fetch a document by its ID or index in dict format or raw string format.
- Return the number of documents in the corpus.
- Check if a given index or ID is in the corpus.
- Return a string representation of the corpus reader.
- Return an iterator for the corpus reader.
- Convert given data into string representation.
- Build an index file for a given data file.
"""

from dataclasses import make_dataclass
import datetime
import json
import mmap
import os

class CorpusReader:
    """A reader for efficiently accessing documents in a corpus using an index.
    """
    def __init__(self, index_path : str):
        """Initialize the corpus reader using an pre-built index file.

        Args:
            index_path (str): Path of the pre-built index file.
        """
        self.index = {}
        self.idx_list = []
        self.idx_set = set()
        self.data_offset = 0
        self.index_path = index_path
        if not os.path.exists(self.index_path):
            raise FileNotFoundError(f'Index file not found: [{self.index_path}]')
        self.index_file_size = os.path.getsize(self.index_path)
        with open(f'{self.index_path}', 'r', encoding='utf-8') as fp:
            for line in fp:
                self.data_offset += len(line.encode())
                idx, start_idx, end_idx = line.strip().split('\t')
                if (idx, start_idx, end_idx) == ('0', '0', '0'):
                    break
                self.index[idx] = {'start': int(start_idx), 'end': int(end_idx)}
                self.idx_list.append(idx)
                self.idx_set.add(idx)
        self.fp = open(f'{self.index_path}', 'r+b')
        self.mm = mmap.mmap(self.fp.fileno(), 0)
        self.data_class = None

    def __del__(self):
        """Clean up resources.
        """
        self.mm.close()
        self.fp.close()

    def __getitem__(self, index : int | str) -> dict:
        """Fetch a document by its ID (str) or index (int) in dict format.

        Args:
            index (int | str): The ID (str) or index (int) of the document.

        Returns:
            dataclass: The fetched document in dataclass format.
        """
        idx = self.idx_list[index] if isinstance(index, int) else index
        if not self.__contain__(idx):
            raise KeyError(f'Index or ID not found: [{idx}]')
        start_idx = self.data_offset+self.index[idx]['start']
        end_idx = self.data_offset+self.index[idx]['end']
        doc = self.mm[start_idx:end_idx]
        doc = doc.decode()
        doc = json.loads(doc)
        if not self.data_class:
            self.data_class = make_dataclass('Document', [(key, type(value)) for key, value in doc.items()])
        instance = self.data_class(**doc)
        return instance

    def raw(self, index : int | str) -> str:
        """Fetch a document by its ID (str) or index (int) in raw string format.

        Args:
            index (int | str): The ID (str) or index (int) of the document.

        Returns:
            str: The fetched document in raw string format.
        """
        idx = self.idx_list[index] if isinstance(index, int) else index
        if not self.__contain__(idx):
            raise KeyError(f'Index or ID not found: [{idx}]')
        start_idx = self.data_offset+self.index[idx]['start']
        end_idx = self.data_offset+self.index[idx]['end']
        doc = self.mm[start_idx:end_idx]
        doc = doc.decode()
        return doc

    def __len__(self) -> int:
        """Return the number of documents in the corpus.

        Returns:
            int: Number of documents in the corpus.
        """
        return len(self.idx_list)

    def __contain__(self, index : int | str) -> bool:
        """Check if the given index (int) or ID (str) is in the corpus.

        Args:
            index (int | str): The ID (str) or index (int) of the document.

        Returns:
            bool: Whether the given index or ID is in the corpus or not.
        """
        return index in self.idx_set

    def __str__(self):
        """Return the string representation of the corpus reader.

        Returns:
            str: String representation of the corpus reader.
        """
        index_file_size = os.path.getsize(self.index_path)
        unit = 'B'
        for current_unit in ['B', 'KB', 'MB', 'GB', 'TB']:
            if index_file_size < 1024:
                break
            index_file_size /= 1024
            unit = current_unit
        return f"CorpusReader(index_path='{self.index_path}')\n- number of documents: {len(self)}\n- index file size: {index_file_size:.1f} {unit}"

    def __repr__(self):
        """Return the string representation of the corpus reader.

        Returns:
            str: String representation of the corpus reader.
        """
        return f"CorpusReader('{self.index_path}')"

    def __iter__(self):
        """Return an iterator for the corpus reader.

        Returns:
            iter: An iterator for the corpus reader.
        """
        for idx in self.idx_list:
            yield self[idx]

    @staticmethod
    def to_str(data : any) -> str:
        """Convert the given data into string representation.

        Args:
            data (any): Data to convert.

        Returns:
            str: String converted version of the given data.
        """
        if isinstance(data, str):
            return data
        data = json.dumps(data, ensure_ascii=False, separators=(',', ':'))
        return data

    @staticmethod
    def build_index(data_path : str, index_path : str, idx_field : str = 'id', verbose : bool = False):
        """Build an index file for the given data file.

        Args:
            data_path (str): Path of the data file (jsonl type).
            index_path (str): Path of the index file to be created.
            idx_field (str): Field of the data file to be used as the ID (str) of the documents. Defaults to 'id'.
            verbose (bool, optional): Wheather to print the progress status or not. Defaults to False.
        """
        # Create index and data files
        with open(f'{index_path}', 'w', encoding='utf-8', newline='\n') as fp_idx, \
            open(f'{index_path}.data', 'w', encoding='utf-8', newline='\n') as fp_data, \
            open(data_path, 'r', encoding='utf-8') as fp:
            file_name = data_path.split('/')[-1]
            cnt = 0
            cnt_doc = 0
            for line in fp:
                data = json.loads(line)
                idx = data[idx_field]
                content = json.dumps(data, ensure_ascii=False, separators=(',', ':'))
                fp_data.write(content)
                size = len(content.encode())
                fp_idx.write(f'{idx}\t{cnt}\t{cnt + size}\n')
                cnt += size
                cnt_doc += 1
                if verbose and cnt_doc % 1000 == 0:
                    now = datetime.datetime.now().strftime('%H:%M:%S')
                    print(f'[ {now} ] Corpus Reader | file: {file_name} | reading documents | doc: {cnt_doc:,} |', end='\r', flush=True)
            if verbose:
                now = datetime.datetime.now().strftime('%H:%M:%S')
                print(f'[ {now} ] Corpus Reader | file: {file_name} | reading documents | doc: {cnt_doc:,} |', flush=True)
            fp_idx.write('0\t0\t0\n')

        # Merge index and data files into one
        with open(f'{index_path}', 'a', encoding='utf-8') as fp_idx, open(f'{index_path}.data', 'r', encoding='utf-8') as fp_data:
            for line in fp_data:
                fp_idx.write(line)
                if verbose and cnt_doc % 1000 == 0:
                    now = datetime.datetime.now().strftime('%H:%M:%S')
                    print(f'[ {now} ] Corpus Reader | file: {file_name} | merging index |', end='\r', flush=True)
            if verbose:
                now = datetime.datetime.now().strftime('%H:%M:%S')
                print(f'[ {now} ] Corpus Reader | file: {file_name} | merging index |', flush=True)
        os.remove(f'{index_path}.data')



from .data_loader_base import DataLoaderQDRBase
from dataclasses import dataclass
from typing import Iterator

@dataclass
class BioASQQueryInstance:
    id : str
    text : str

@dataclass
class BioASQDocInstance:
    id : str
    text : str
    title : str
    journal : str
    year : int
    pmid : str
    mesh_major : list[str]

@dataclass
class BioASQQrelInstance:
    qid : str
    did : str
    score : int

class BioASQDataLoader(DataLoaderQDRBase):
    def __init__(self, *args, **kwargs):
        super().__init__('beir/bioasq', *args, **kwargs)

    def make_query_instance(self, **kwargs) -> BioASQQueryInstance:
        return BioASQQueryInstance(**kwargs)

    def make_doc_instance(self, **kwargs) -> BioASQDocInstance:
        return BioASQDocInstance(**kwargs)

    def make_qrel_instance(self, **kwargs) -> BioASQQrelInstance:
        return BioASQQrelInstance(**kwargs)

    def __str__(self) -> str:
        return 'BioASQDataLoader()'

    def __repr__(self) -> str:
        return 'BioASQDataLoader()'



from setuptools import setup, find_packages

setup(
    name='hamu-tool',
    version='0.0.14',
    description='Many useful tools for computer scientists!',
    license='MIT',
    author='DoHyeon Lee',
    author_email='waylight3@snu.ac.kr',
    url='https://github.com/waylight3/hamu-tool',
    keywords=['hamu', 'tool', 'computer', 'science', 'useful', 'toolkit', 'library', 'python', 'package', 'module', 'utility', 'function', 'class', 'method'],
    install_requires=[],
    python_requires='>=3.11',
    packages=find_packages(),
    zip_safe=False
)



def is_int(x : str) -> bool:
    """Check if a string is an integer.

    Args:
        x (str): The string to check.

    Returns:
        bool: True if the string is an integer, False otherwise.
    """
    try:
        int(x)
        return True
    except ValueError:
        return False

def is_float(x : str) -> bool:
    """Check if a string is a float.

    Args:
        x (str): The string to check.

    Returns:
        bool: True if the string is a float, False otherwise.
    """
    try:
        float(x)
        return True
    except ValueError:
        return False

def is_number(x : str) -> bool:
    """Check if a string is a number.

    Args:
        x (str): The string to check.

    Returns:
        bool: True if the string is a number, False otherwise.
    """
    return is_int(x) or is_float(x)

def mean(data : list) -> float:
    """Calculate the mean of a list of numbers.

    Args:
        data (list): A list of numbers.

    Returns:
        float: The mean of the numbers.
    """
    return sum(data) / len(data)

def variance(data : list) -> float:
    """Calculate the variance of a list of numbers.

    Args:
        data (list): A list of numbers.

    Returns:
        float: The variance of the numbers.
    """
    m = mean(data)
    return mean([(x - m) ** 2 for x in data])

def std_dev(data : list) -> float:
    """Calculate the standard deviation of a list of numbers.

    Args:
        data (list): A list of numbers.

    Returns:
        float: The standard deviation of the numbers.
    """
    return variance(data) ** 0.5

def median(data : list) -> float:
    """Calculate the median of a list of numbers.

    Args:
        data (list): A list of numbers.

    Returns:
        float: The median of the numbers.
    """
    data = sorted(data)
    n = len(data)
    if n % 2 == 1:
        return data[n // 2]
    else:
        return (data[n // 2 - 1] + data[n // 2]) / 2

